# -*- coding: utf-8 -*-
"""
Copyright (C) 2014 Benjamin Eltzner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 3
of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
or see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import numpy.linalg as la
import math
from math import sqrt, cos
from functools import reduce
from operator import mul
from scipy.integrate import quad
from scipy.stats import gmean
from RESH_io import import_csv, export_csv, find_files, crop_name
from RESH_plot import (circle_plot, sphere_views, line_plot, var_plot,
                       residual_plots, rainbow, make_circle, scatter_2D)
from RESH_PNS import pns_loop, unfold_points
from RESH_geometry import (full_RESHify, RESHify_with_radii,
                           unRESHify, mean_on_sphere, tangent_space)

################################################################################
################################   Constants   #################################
################################################################################

PI = np.pi
DEG = math.degrees(1)
RAD = math.radians(1)
EPS = 1e-10

"""#############################################################################
Supporting function
#############################################################################"""

def travel (p1, p2, resh = True):
    if resh:
        return np.arccos(np.einsum('ij,ij->i', p1, p2).clip(-1,1)) * DEG
    return la.norm(np.array([np.arccos(np.einsum('ij,ij->i', p1[k], p2[k]).clip(-1,1))
                    for k in range(len(p1))]), axis=0) * DEG

def maximal_residuals (n):
    res_max = []
    res_max.append(sqrt(PI**2/3)*DEG)
    for k in range(1,n):
        def f(x): return x**2 * cos(x)**k
        def g(x): return cos(x)**k
        res_max.append(sqrt(quad(f,0,0.5*PI)[0]/quad(g,0,0.5*PI)[0])*DEG)
    return np.array(res_max)

def import_data (file, hippo_count, suffix=''):
    name = crop_name(file) + suffix
    print(name)
    if 'hippo' in name:
        data = import_csv(file)
        tmp = data['spoke_orientations']
        colors = [rainbow(i, len(tmp)) for i in range(len(tmp))]
        if hippo_count == 0:
            name += '_all'
            radii = np.hstack((data['spoke_lengths'], data['shape_length']))
            spheres = ([tmp[:,i,:] for i in range(tmp.shape[1])] +
                       [data['shape_orientations']])
            cpns_results = np.vstack((data['cpns_shape'], data['cpns_spokes']))
            hippo_count += 1
        else:
            name += '_spokes'
            radii = data['spoke_lengths']
            spheres = [tmp[:,i,:] for i in range(tmp.shape[1])]
            cpns_results = data['cpns_spokes']
    else:
        data = import_csv(file)
        spheres = data['s_reps']
        cpns_results = data['cpns_results']
        colors = [rainbow(i, len(spheres)) for i in range(len(spheres))]
        if 'matlab_d1' in name:
            order = [20,0,7,26,9,6,4,11,29,14,5,25,22,2,28,3,19,24,21,23,27,17,15,16,12,10,8,18,13,1]
            spheres = spheres[order]
            cpns_results = cpns_results[:,order]
        if 'matlab_d2' in name:
            order = [20,2,26,3,23,14,25,11,1,17,5,6,18,15,27,8,4,19,12,10,21,29,22,9,0,24,28,7,13,16]
            spheres = spheres[order]
            cpns_results = cpns_results[:,order]
        spheres = [spheres[:,i,:] for i in range(spheres.shape[1])]
        radii = np.ones(len(spheres))
    return name, spheres, radii, colors, hippo_count, cpns_results

def make_line_plots (distances, colors, name):
    data = []
    for i, d in enumerate(distances):
        max_dim = min(d.shape)
        val, vec = la.eig(np.cov(d))
        val = val.real; vec = vec.real
        d = np.einsum('ji,jk->ki', d, vec)[::-1]
        tmp = '_PDNS' if i == 0 else '_CPNS'
        #residual_plots(d, None, 3, colors, name + tmp + '_residuals')
        val = np.array([list(range(1, len(val)+1)), val]).T
        val[:,1] *= 100 / np.sum(val[:,1])
        tmp=1
        while val[:tmp,1].sum() < 90:
            tmp+=1
        print(val[:tmp,1], val[:tmp,1].sum())
        data.append(val[:max_dim])
    var_plot(data, 90, name + '_variances')

def many_plots (spheres, projected_points, stdev, dimensions, colors, name):
    circle = make_circle(spheres[-2].normal, spheres[-2].height)
    sphere_views(projected_points[-3], spheres[-2].normal, colors, circle, name + '_resh_sphere')
    scatter_2D(tangent_space(projected_points[-3],
                             mean_on_sphere(projected_points[-3], False)[0]) * DEG,
               colors, name + '_2D_euclidean')
    circle_plot(projected_points[-2], colors, False, name + '_resh_circle')
    line_plot(np.array([list(range(len(stdev))),
                        100*(stdev[::-1]/maximal_residuals(len(stdev)))**2]).T,
              ['Dimension', 'variance relative to uniform [\%]'],
              name + '_relative_variances')
    unfolded_points = [unfold_points(pts, spheres[:i+1])
                       for i, pts in enumerate(projected_points[:-1])]
    unreshed_points = [unRESHify(pts, dimensions) for pts in unfolded_points]
    dists_resh = travel(unfolded_points[0], unfolded_points[-2])
    dists_prod = travel(unreshed_points[0], unreshed_points[-2], False)
    sphere_views(projected_points[-3], spheres[-2].normal, dists_resh, circle, name + '_resh_sphere_heat_resh')
    sphere_views(projected_points[-3], spheres[-2].normal, dists_prod, circle, name + '_resh_sphere_heat_prod')

def unit_scale(data):
    return np.vstack(tuple([d / la.eig(np.cov(d))[0].sum().real for d in data]))

"""#############################################################################
Control functions
#############################################################################"""

def hippocampi ():
    files = find_files('*.csv', 'data/Dagsthul_hippoData')
    for f in files:
        name = crop_name(f) + '_with_radii'
        print(name)
        data = import_csv(f)
        tmp = data['spoke_orientations']
        colors = [rainbow(i, len(tmp)) for i in range(len(tmp))]
        radii = np.hstack((data['spoke_lengths'], data['shape_length']))
        spheres = ([tmp[:,i,:] for i in range(tmp.shape[1])] +
                   [data['shape_orientations']])
        scale_data = data['cpns_radii']
        cpns = unit_scale([np.vstack((data['cpns_shape'], data['cpns_spokes'])),
                           scale_data])
        resh = RESHify_with_radii(spheres, radii, colors)
        print(resh.shape)#, la.norm(resh, axis=1))
        spheres, projected_points, distances = pns_loop(resh)
        print('Radii:\n', [s.flat_radius for s in spheres[:-1]])
        scales = np.array([1] + [reduce(mul, [s.flat_radius
                                 for s in spheres[:i]], 1)
                                 for i in range(1,len(spheres))]) * gmean(radii)
        distances = np.einsum('ij,i->ij', distances, scales) * RAD
        full_data = unit_scale([distances, scale_data])
        make_line_plots([full_data, cpns], colors, name)

def s_reps ():
    files = (find_files('*.csv', 'data/sungkyu_toy_data') +
             find_files('*.csv', 'data/Dagsthul_hippoData')*2)
    stdev_data = {}
    count = 0
    for f in files:
        #if not 'matlab_d1' in f: continue
        name, spheres, radii, colors, count, cpns = import_data(f, count)
        dimensions = [s.shape[1] for s in spheres]
        resh = RESHify_with_radii(spheres, radii, colors)
        print(resh.shape)#, la.norm(resh, axis=1))
        spheres, projected_points, distances = pns_loop(resh)
        stdev = np.array([np.std(d) for d in distances])
        print('Radii:\n', [s.flat_radius for s in spheres[:-1]])
        stdev_data[name] = stdev
        #many_plots(spheres, projected_points, stdev, dimensions, colors, name)
        scales = np.array([1] + [reduce(mul, [s.flat_radius
                                 for s in spheres[:i]], 1)
                                 for i in range(1,len(spheres))])
        distances = np.einsum('ij,i->ij', distances, scales)
        make_line_plots([distances, cpns * DEG], colors, name)
    export_csv(stdev_data, 'stdevs.csv')

def s_reps_cpns ():
    files = (find_files('*.csv', 'data/sungkyu_toy_data') +
             find_files('*.csv', 'data/Dagsthul_hippoData')*2)
    count = 0
    for f in files:
        if not 'hippo' in f: continue
        name, spheres, radii, colors, count, _ = import_data(f, count, '_CPNS')
        all_distances = None
        for i,s in enumerate(spheres):
            spheres, projected_points, distances = pns_loop(s, mode='scale')
            distances = np.array(distances) * radii[i]
            for j, sp in enumerate(spheres[:-1]):
                distances[i+1:] *= sp.flat_radius
            print('Sphere', (i+1), 'Radii:', [s.flat_radius for s in spheres[:-1]], distances.shape)
            if all_distances is None:
                all_distances = distances
            else:
                all_distances = np.vstack((all_distances, distances))
        max_dim = min(all_distances.shape)
        all_distances = all_distances.T * RAD# euclideanize(all_distances * RAD).T
        val, vec = la.eig(np.cov(all_distances.T))
        val = val.real; vec = vec.real
        all_distances = np.einsum('ij,jk->ki', all_distances, vec)[::-1]
        val = np.array([list(range(1, len(val)+1)), val]).T
        val[:,1] *= 100 / np.sum(val[:,1])
        line_plot(val[:max_dim], ['codimension', 'variances [$\%$]'], name + '_variances')
        residual_plots(all_distances, np.max(np.abs(all_distances))/160., 6, colors, name + '_residuals')

def s_rep_tests ():
    files = find_files('*.csv', 'data/sungkyu_toy_data')
    files = [f for f in files if 'matlab' in f]
    for f in files:
        name = crop_name(f)
        print(name)
        data = import_csv(f)['s_reps']
        print(data.shape)
        flat_data_in = np.array([data[i].flatten() for i in range(data.shape[0])])
        print(np.sqrt(la.eig(np.cov(flat_data_in.T))[0][:5]))
        colors = [rainbow(i, len(data)) for i in range(len(data))]
        resh = full_RESHify([data[:,i,:] for i in range(data.shape[1])], colors)
        print(np.sqrt(la.eig(np.cov(resh.T))[0][:5]))
        new_data = unRESHify(resh)
        print(new_data.shape)
        flat_new_data = np.array([new_data[i].flatten() for i in range(new_data.shape[0])])
        print(np.sqrt(la.eig(np.cov(flat_new_data.T))[0][:5]))
#        spheres, points, _ = pns_loop(resh)
#        for s in spheres:
#            print(s.flat_radius)
#        for p in points:
#            print(np.sqrt(la.eig(np.cov(p.T))[0][:3]))
#        for i in range(data.shape[1]):
#            sphere = data[:,i,:]
#            sphere_plot(sphere, colors)


if __name__ == "__main__":
    #hippocampi()
    s_reps()
    #s_reps_cpns()
    #s_rep_tests()
